#ifndef VERSION_H
#define VERSION_H


/* Program's version number. */
#define IMAPFILTER_VERSION	"2.2.2"

/* Program's copyright. */
#define IMAPFILTER_COPYRIGHT	"Copyright (c) 2001-2010 Eleftherios Chatzimparmpas"


#endif				/* VERSION_H */
